package iSATLibrary;

import com.sun.jna.Callback;
import com.sun.jna.Library;
import com.sun.jna.Native;
import com.sun.jna.Pointer;
import com.sun.jna.ptr.IntByReference;

/**
 * This class is the library to use incremental SAT APIs through JNA.
 * 
 * @author Tatsuya Sako
 */
public class ISatLibrary {

	protected String solverName;
	protected IPASIR ipasirLib;
	protected ISATLIB isatLib;
	protected Pointer solverPtr;

	// constructor
	public ISatLibrary(String name) {
		solverName = name;
		ipasirLib = (IPASIR) Native.loadLibrary(solverName, IPASIR.class);
		isatLib = (ISATLIB) Native.loadLibrary(solverName, ISATLIB.class);
		solverPtr = ipasirLib.ipasir_init();
	}

	// IPASIR API (by SAT Race 2015)
	public interface IPASIR extends Library {

		public String ipasir_signature();
		public Pointer ipasir_init();
		public void ipasir_release(Pointer solver);
		public void ipasir_add(Pointer solver, int lit_or_zero);
		public void ipasir_assume(Pointer solver, int lit);
		public int ipasir_solve(Pointer solver);
		public int ipasir_val(Pointer solver, int lit);
		public int ipasir_failed(Pointer solver, int lit);
		public void ipasir_set_terminate(Pointer solver, Pointer state, Callback callback);

	}

	// proposed API
	public interface ISATLIB extends Library {

		/**
		 * Sets the verbosity of SAT solver's output.
		 * 
		 * @param solver
		 *	the pointer to SAT solver
		 * @param v
		 *  verbosity
		 */
		public void isat_set_verbosity(Pointer solver, int v);

		/**
		 * Gets the number of variables of SAT solver.
		 * 
		 * @param solver
		 *  the pointer to SAT solver
		 * @return
		 *  number of variables
		 */
		public int isat_vars(Pointer solver);

		/**
		 * Adds clause(s) to SAT solver. 
		 * If you add some clauses, you have to separate each of them by "0".
		 * 
		 * @param solver
		 *  the pointer to SAT solver
		 * @param clauses
		 *  the array of clauses
		 * @param length
		 *  the length of clauses array
		 */
		public void isat_add_clauses(Pointer solver, int[] clauses, int length);

		/**
		 * Gets assignments of all variables.
		 * 
		 * @param solver
		 *  the pointer to SAT solver
		 * @return
		 *  the pointer to the head of model array
		 */
		public IntByReference isat_model(Pointer solver);

		/**
		 * Releases a pointer allocated native code.
		 * 
		 * @param solver
		 *  the pointer to SAT solver
		 * @param p
		 *  the pointer required to release
		 */
		public void isat_clean(Pointer solver, Pointer p);

		/**
		 * Gets elapsed time from invocation of SAT solver.
		 * 
		 * @param solver
		 *  the pointer to SAT solver
		 * @return
		 *  elapsed time
		 */
		public double isat_get_time(Pointer solver);

		/**
		 * Adds blocking clause to restrict a certain solution.
		 * 
		 * @param solver
		 *  the pointer to SAT solver
		 * @param literals
		 *  literals
		 * @param length
		 *  the length of literals array
		 */
		public void isat_add_blocking_clause(Pointer solver, int[] literals, int length);

		/**
		 * Freezes a literal. Frozen literal is not simplified in SAT solver.
		 * This method is available for only "simp" solver.
		 * 
		 * @param solver
		 *  the pointer to SAT solver
		 * @param literal
		 *  the literal required to freeze
		 */
		public void isat_freeze(Pointer solver, int literal);

		/**
		 * Checks if literal is frozen. 
		 * This method is available for only "simp" solver.
		 * 
		 * @param solver
		 *  the pointer to SAT solver
		 * @param literal
		 *  the literal required to check if frozen
		 * @return 
		 *  true (if the literal is frozen), false (otherwise)
		 */
		public boolean isat_is_eliminated(Pointer solver, int literal);

		// unimplemented
		// public void isat_addPB(Pointer solver, int[] coefficients, int[] literals);
		// public IntByReference isat_MUS(Pointer solver);
		// public void isat_delete_learnts(Pointer solver);
		// public void isat_set_polarity(Pointer solver, int[] polarity);
		// public IntByReference isat_polarity(Pointer solver);
	}

}
