package iSugar;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.converter.Converter;
//import jp.kobe_u.sugar.csp.BooleanVariable;
import jp.kobe_u.sugar.csp.CSP;
//import jp.kobe_u.sugar.csp.Clause;
import jp.kobe_u.sugar.csp.IntegerDomain;
import jp.kobe_u.sugar.csp.IntegerVariable;
import jp.kobe_u.sugar.encoder.Encoder;
import jp.kobe_u.sugar.expression.Expression;
import static iSugar.Solver.Status.*;

public class Graph {
	
	private CSP csp;
	private Solver satSolver;
	private static int countOfIterations = 0;
	private static int countOfCycles = 0;
	
	public static TreeSet<Integer> nodes;
	public static TreeSet<int[]> edges;
	
	private static Map<Integer, Set<Integer>> adjacentMap;
	
	private static int counter = 0;
	
	double start, stop, encodetime, solvetime;
	
	// constructor
	 public Graph(Solver satSolver){
		this.csp = new CSP();
		this.satSolver = satSolver;
		nodes = new TreeSet<Integer>();
		edges = new TreeSet<int[]>(new Comparator<int[]>(){
			@Override
			public int compare(int[] a0, int[] a1) {
				if (a0[0] != a1[0])
					return a0[0] < a1[0] ? -1 : 1;
				if (a0[1] != a1[1])
					return a0[1] < a1[1] ? -1 : 1;
				return 0;
			}});
		adjacentMap = new HashMap<Integer, Set<Integer>>();
	}
	
	public void setCSP(CSP csp){
		this.csp = csp;
	}
	
	// generate edges
	public int[] edge(int n1, int n2) {
		int[] edge = new int[2];
		if(n1 < n2){
			edge[0] = n1; edge[1] = n2;
		}
		else{
			edge[0] = n2; edge[1] = n1;
		}
		return edge;	
	}
	
	// add a condition, "n1 have n2 as a set of adjacent nodes"
	private void addAdjacent(int n1, int n2){
		if(adjacentMap.containsKey(n1))
			adjacentMap.get(n1).add(n2); // add to set of adjacent nodes
		else{
         Set<Integer> set = new LinkedHashSet<Integer>(); // generate set of adjacent nodes
			set.add(n2);
			adjacentMap.put(n1, set); // add to adjacent map
		}
	}
	
	// add a vertex to graph
	private void addNode(int n){
		nodes.add(n);
	}

	// add an edge to graph
	private void addEdge(int n1, int n2){
		if(n1 != n2){
			int[] edge = edge(n1,n2);
				edges.add(edge);
				addAdjacent(n1, n2);
				addAdjacent(n2, n1);				
		}	
	}	

	// get the set of adjacent nodes
	private Set<Integer> adjacent(int n){
		return adjacentMap.get(n);
	}
	
	// load the file and add the information to the graph object
	public void parse(String fileName){
		try {
			BufferedReader rd = new BufferedReader(new FileReader(fileName));
			String line;
			String regex = "e\\s+(\\d+)\\s+(\\d+)";
			Pattern p = Pattern.compile(regex);
			Matcher m;
			// ignore except the line "e (num) (num)"
			while(true){
				line = rd.readLine();
				if(line == null) break;
				if(!line.trim().startsWith("e")) continue;
				m = p.matcher(line);
				if(m.find()){
					int n1 = Integer.parseInt(m.group(1));
					addNode(n1);
					int n2 = Integer.parseInt(m.group(2));
					addNode(n2);
					addEdge(n1, n2);
				}	
			}
			rd.close();
		} catch (FileNotFoundException e) {
			System.out.println("File Not Found.");
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	// get a cycle from SAT solution
	public List<Integer> getCycle(int node, int initial, List<Integer> cycle){
		int node2 = getNextNode(node);
		cycle.add(0, node2);
		if(node2 != initial){
			cycle = getCycle(node2, initial, cycle);
		}
		return cycle;
	}
	
	// get the next node adjacent to a node
	private int getNextNode(int node) {
		//System.out.println("node : "+node);
		Set<Integer> adjacentNodes = adjacent(node);
		for(int n : adjacentNodes){
			//System.out.print("a_"+node+"_"+n + " ");
			int a = csp.getIntegerVariable("a_"+node+"_"+n).getValue();
			if(a>0){
				//System.out.println(a+" hoge");
				return n;
			}
		}
		System.out.println("ERROR");
		System.exit(1);
		return 0;
	}

	// get all cycles
	public Set<List<Integer>> getCycles(){
		Set<List<Integer>> cycles = new LinkedHashSet<List<Integer>>();
		List<Integer> nodes1 = new ArrayList<Integer>(nodes);
		while(!nodes1.isEmpty()){
			int node = nodes1.get(0);
			List<Integer> cycle = new ArrayList<Integer>();
			cycle.add(node);
			cycle = getCycle(node, node, cycle);
			cycles.add(cycle);
			nodes1.removeAll(cycle);
		}
 		return cycles;
	}
	
	// generate a CSP
	public void define() throws SugarException {
		IntegerDomain domain = IntegerDomain.create(0, 1);
		String name;
		List<Expression> allExpressions = new ArrayList<Expression>();
		// definition the vertices and constraint of edges
		for(int[] edge : edges){
			List<Expression> expressions = new ArrayList<Expression>();
			List<Expression> innerExpressions = new ArrayList<Expression>();
			expressions.add(Expression.LE); // [<= ]
			innerExpressions.add(Expression.ADD); // [+ ]
			name = "a_"+edge[0]+"_"+edge[1];
			try{
				IntegerVariable v1;
				v1 = new IntegerVariable(name, domain);
				csp.add(v1);
				innerExpressions.add(Expression.create(name)); // [+ a_ij]
				name = "a_"+edge[1]+"_"+edge[0];
				IntegerVariable v2 = new IntegerVariable(name, domain);
				csp.add(v2);
				innerExpressions.add(Expression.create(name)); // [+ a_ij a_ji]
				expressions.add(Expression.create(innerExpressions)); // [<= [+ a_ij a_ji]]
				expressions.add(Expression.create(1)); // [<= [+ a_ij a_ji] 1]  a_ij + a_ji <= 1 
				allExpressions.add(Expression.create(expressions));
			} catch(SugarException e){
				continue;
			}
		}
		// constraint of in and out degree
		int listSize; // >=2
		for(int n1: nodes){
			List<Integer> adjacentList = new ArrayList<Integer>(adjacent(n1));			
			listSize = adjacentList.size();
			IntegerVariable s;
			List<Expression> expressions_in = new ArrayList<Expression>();
			List<Expression> innerExpressions_in = new ArrayList<Expression>();
			List<Expression> expressions_out = new ArrayList<Expression>();
			List<Expression> innerExpressions_out = new ArrayList<Expression>();

			// UNSAT if size of the list is 1 (cannot generate any cycle)
			if(listSize < 2){
				System.out.println("This Graph has a node that has only zero or one adjacent node");
				System.out.println("This Graph has no Hamiltonian Cycle");
				System.exit(0);
			}
			
			// size = 2
			else if(listSize == 2){
				// in degree
				expressions_in.add(Expression.EQ); // [= ]
				innerExpressions_in.add(Expression.ADD); // [+ ]			
				name = "a_"+adjacentList.get(0)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1]
				name = "a_"+adjacentList.get(1)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1 a_(i+1)n1]			
				expressions_in.add(Expression.create(new ArrayList<Expression>(innerExpressions_in))); // [= [+ a_in1 a_(i+1)n1]]
				innerExpressions_in.clear();
				expressions_in.add(Expression.create(1)); // [= [+ a_in1 a_(i+1)n1] 1]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_in))); // 制約群に追加
				expressions_in.clear();
				// out degree
				expressions_out.add(Expression.EQ); // [= ]
				innerExpressions_out.add(Expression.ADD); // [+ ]			
				name = "a_"+n1+"_"+adjacentList.get(0);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i]
				name = "a_"+n1+"_"+adjacentList.get(1);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i a_n1(i+1)]			
				expressions_out.add(Expression.create(new ArrayList<Expression>(innerExpressions_out))); // [= [+ a_n1i a_n1(i+1)]]
				innerExpressions_out.clear();
				expressions_out.add(Expression.create(1)); // [= [+ a_n1i a_n1(i+1)] 1]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_out))); // 制約群に追加
				expressions_out.clear();
			}
			// introduce complement variable if size >= 3
			else{
				// in degree
				// s = x1 + x2
				expressions_in.add(Expression.EQ); // [= ]
				innerExpressions_in.add(Expression.ADD); // [+ ]			
				name = "a_"+adjacentList.get(0)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1]
				name = "a_"+adjacentList.get(1)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1 a_(i+1)n1]			
				expressions_in.add(Expression.create(new ArrayList<Expression>(innerExpressions_in))); // [= [+ a_in1 a_(i+1)n1]]
				innerExpressions_in.clear();
				name = "s_"+counter;
				//counter++;
				s = new IntegerVariable(name, domain);
				csp.add(s);
				expressions_in.add(Expression.create(name)); // [= [+ a_in1 a_(i+1)n1] s]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_in))); // 制約群に追加
				expressions_in.clear();
				// s' = x3 + s
				for(int i=2; i<listSize-1; i++){
					expressions_in.add(Expression.EQ); // [= ]
					innerExpressions_in.add(Expression.ADD); // [+ ]			
					name = "a_"+adjacentList.get(i)+"_"+n1;
					innerExpressions_in.add(Expression.create(name)); // [+ a_in1]
					name = "s_"+counter;
					counter++;
					innerExpressions_in.add(Expression.create(name)); // [+ a_in1 s]
					expressions_in.add(Expression.create(new ArrayList<Expression>(innerExpressions_in))); // [= [+ a_in1 s]]
					innerExpressions_in.clear();
					name = "s_"+counter;
					IntegerVariable s1 = new IntegerVariable(name, domain);
					csp.add(s1);
					expressions_in.add(Expression.create(name)); // [= [+ a_in1 s] s1]
					allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_in))); // 制約群に追加
					expressions_in.clear();
				}	
				// 1 = xn + s''
				expressions_in.add(Expression.EQ); // [= ]
				innerExpressions_in.add(Expression.ADD); // [+ ]			
				name = "a_"+adjacentList.get(listSize-1)+"_"+n1;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1]
				name = "s_"+counter;
				counter++;
				innerExpressions_in.add(Expression.create(name)); // [+ a_in1 s]
				expressions_in.add(Expression.create(new ArrayList<Expression>(innerExpressions_in))); // [= [+ a_in1 s]]
				innerExpressions_in.clear();
				expressions_in.add(Expression.create(1)); // [= [+ a_in1 s] 1]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_in))); // 制約群に追加
				expressions_in.clear();
				// out degree
				// s = x1 + x2
				expressions_out.add(Expression.EQ); // [= ]
				innerExpressions_out.add(Expression.ADD); // [+ ]			
				name = "a_"+n1+"_"+adjacentList.get(0);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i]
				name = "a_"+n1+"_"+adjacentList.get(1);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i a_n1(i+1)]			
				expressions_out.add(Expression.create(new ArrayList<Expression>(innerExpressions_out))); // [= [+ a_n1i a_n1(i+1)]]
				innerExpressions_out.clear();
				name = "s_"+counter;
				//counter++;
				s = new IntegerVariable(name, domain);
				csp.add(s);
				expressions_out.add(Expression.create(name)); // [= [+ a_n1i a_n1(i+1)] s]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_out))); // 制約群に追加
				expressions_out.clear();
				// s' = x3 + s
				for(int i=2; i<listSize-1; i++){
					expressions_out.add(Expression.EQ); // [= ]
					innerExpressions_out.add(Expression.ADD); // [+ ]			
					name = "a_"+n1+"_"+adjacentList.get(i);
					innerExpressions_out.add(Expression.create(name)); // [+ a_n1i]
					name = "s_"+counter;
					counter++;
					innerExpressions_out.add(Expression.create(name)); // [+ a_n1i s]
					expressions_out.add(Expression.create(new ArrayList<Expression>(innerExpressions_out))); // [= [+ a_n1i s]]
					innerExpressions_out.clear();
					name = "s_"+counter;
					IntegerVariable s1 = new IntegerVariable(name, domain);
					csp.add(s1);
					expressions_out.add(Expression.create(name)); // [= [+ a_n1i s] s1]
					allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_out))); // 制約群に追加
					expressions_out.clear();
				}	
				// 1 = xn + s''
				expressions_out.add(Expression.EQ); // [= ]
				innerExpressions_out.add(Expression.ADD); // [+ ]			
				name = "a_"+n1+"_"+adjacentList.get(listSize-1);
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i]
				name = "s_"+counter;
				counter++;
				innerExpressions_out.add(Expression.create(name)); // [+ a_n1i s]
				expressions_out.add(Expression.create(new ArrayList<Expression>(innerExpressions_out))); // [= [+ a_n1i s]]
				innerExpressions_out.clear();
				expressions_out.add(Expression.create(1)); // [= [+ a_n1i s] 1]
				allExpressions.add(Expression.create(new ArrayList<Expression>(expressions_out))); // 制約群に追加
				expressions_out.clear();

			}
		}

		Converter converter = new Converter(csp);
		converter.convert(allExpressions);
		//csp.output(System.out, ""); // for debug
		converter = null;
		allExpressions = null;
		Expression.clear();
		Runtime.getRuntime().gc();
	}

	// add blocking clauses that is not Hamiltonian cycle
	public void addBlockingClauses(List<Integer> cycle) throws SugarException{
		int size = cycle.size();
		//System.out.println("size : "+size);
		int n1, n2;
		int[] blockingClause1 = new int[size-1], blockingClause2 = new int[size-1];
		for(int i = 0; i<size-1; i++){
			n1 = cycle.get(i); n2 = cycle.get(i+1);
			blockingClause1[i] = csp.getIntegerVariable("a_"+n1+"_"+n2).getCode();
			blockingClause2[i] = csp.getIntegerVariable("a_"+n2+"_"+n1).getCode();
		}
		satSolver.addNormalizedClause(blockingClause1);
		satSolver.addNormalizedClause(blockingClause2);
	}

	// add blocking clauses that is not Hamiltonian cycle(not use incremental method)
	 public void addBlockingClauses(List<Integer> cycle, String satFileName) throws SugarException, IOException{
		int size = cycle.size();
		//System.out.println("size : "+size);
		int n1, n2;
		ArrayList<Integer> blockingClause1 = new ArrayList<Integer>(), blockingClause2 = new ArrayList<Integer>();
		for(int i = 0; i<size-1; i++){
			n1 = cycle.get(i); n2 = cycle.get(i+1);
			blockingClause1.add(csp.getIntegerVariable("a_"+n1+"_"+n2).getCode());
			blockingClause2.add(csp.getIntegerVariable("a_"+n2+"_"+n1).getCode());
		}
		// write to csp file
		StringBuilder sb = new StringBuilder("");
		File file = new File(satFileName);
		BufferedReader br = new BufferedReader(new FileReader(file));
		Pattern p = Pattern.compile("(p cnf \\d+ )(\\d+)");
		Matcher m;
			 String line = br.readLine();
			 m = p.matcher(line);
		if(m.find()){
			 int clauses = Integer.parseInt(m.group(2)) + 2;
			 sb.append(m.group(1)+clauses+"\n");
		}
		while((line=br.readLine()) != null){
			 sb.append(line+"\n");
		}
		br.close();
		FileWriter filewriter = new FileWriter(file);
		filewriter.write(sb.toString());
		for(int i: blockingClause1){
			 filewriter.write(i+" ");
		}
		filewriter.write("0\n");
		for(int i: blockingClause2){
			 filewriter.write(i+" ");
		}
		filewriter.write("0\n");
		filewriter.close();
	}
	
	// solve Hamiltonian cycle problem
	public void solveHCP() throws SugarException{
		System.out.println("c Nodes : "+nodes.size());
		System.out.println("c Edges : "+edges.size());
		System.out.println();
		System.out.println("c Encoding...");
		start = System.nanoTime();
		Encoder encoder = new Encoder(csp);
		encoder.encode(satSolver);
		stop = System.nanoTime();
		encodetime = (stop-start) / 1000000000.0;
		System.out.println("c Finished Encoding.\n");
		System.out.printf("c Encoding Time : %f sec\n", encodetime);
		//System.out.printf("Encode Time : %f sec\n", SatSolver4JNA.addclausetime);
		System.out.printf("c variables : %d\nc clauses : %d\n", satSolver.variablesCount, satSolver.clausesCount);

		start = System.nanoTime();
		System.out.println("c Searching Hamiltonian Cycle...\n");
		while(satSolver.solve()==SATISFIABLE){
			countOfIterations++;
			/*
			for (IntegerVariable v : csp.getIntegerVariables()) {
				System.out.println("code of "+v.getName()+" = "+v.getCode());
			}
			satSolver.outputModel();
			*/
			BitSet model = satSolver.getModel();
			for (IntegerVariable v : csp.getIntegerVariables()) {
				v.decode(model);
				//System.out.println(v.getName()+" = "+v.getValue());
			}
			// get cycles
			Set<List<Integer>> cycles = getCycles();
			if(cycles.size()==1){ // if only one cycle, solution found
				countOfCycles++;
				System.out.println("c A Hamiltonian Cycle is Found");
				System.out.print("a ");
				for(List<Integer> cycle : cycles){
					for(int node : cycle){
						System.out.print(node+" ");
					}
				}
				System.out.println("\nc Iteration : " + countOfIterations);
				System.out.println("c Cycles : "+ countOfCycles);
				stop = System.nanoTime();
				solvetime = (stop-start) / 1000000000.0;
				System.out.printf("c Solving Time : %f sec\n", solvetime);
				System.out.printf("c Average solving Time : %f sec\n", solvetime/(countOfIterations+1));
				System.exit(0);
			}
			else{ // if more than one cycle, not solution
				System.out.println("c "+cycles.size()+" cycles found");
				countOfCycles += cycles.size();
				for(List<Integer> cycle : cycles)
					addBlockingClauses(cycle);
			}
		}
		System.out.println("c This Graph has no Hamiltonian Cycle");
		System.out.println("c Iteration : " + countOfIterations);
		System.out.println("c Cycles : "+ countOfCycles);
		stop = System.nanoTime();
		solvetime = (stop-start) / 1000000000.0;
		System.out.printf("c Solving Time : %f sec\n", solvetime);
		System.out.printf("c Average solving Time : %f sec\n", solvetime/(countOfIterations+1));
	}

	 // solve Hamiltonian cycle problem
	 public void solveHCP(String solverName) throws SugarException, IOException{
		String satFileName = "tmp.cnf";
		String mapFileName = "tmp.map";
		String outFileName = "tmp.out";
		System.out.println("c Nodes : "+nodes.size());
		System.out.println("c Edges : "+edges.size());
		System.out.println();
		System.out.println("c Encoding...");
		start = System.nanoTime();
		Encoder encoder = new Encoder(csp);
		encoder.encode(satFileName);
		encoder.outputMap(mapFileName);
		stop = System.nanoTime();
		encodetime = (stop-start) / 1000000000.0;
		System.out.println("c Finished Encoding.\n");
		System.out.printf("c Encoding Time : %f sec\n", encodetime);
		//System.out.printf("Encode Time : %f sec\n", SatSolver4JNA.addclausetime);
		System.out.printf("c variables : %d\nc clauses : %d\n", satSolver.variablesCount, satSolver.clausesCount);

		start = System.nanoTime();
		System.out.println("c Searching Hamiltonian Cycle...\n");
		// invoke a SAT solver as an another process
		ProcessBuilder pb = new ProcessBuilder(solverName, satFileName, outFileName);
		while(true){
			countOfIterations++;
			pb.redirectErrorStream(true);
			try {
				 Process p = pb.start();	
				 p.waitFor();
				 System.out.println(pb.redirectInput());
				 try (BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()))) {
					 for(String line = br.readLine(); line != null; line = br.readLine())
						 System.out.println(line);
				 }
			} catch (IOException | InterruptedException e) {
				 e.printStackTrace();
			}

			// decode
			if(encoder.decode(outFileName)){
				 // get cycles
				 Set<List<Integer>> cycles = getCycles();
				 if(cycles.size()==1){
					  countOfCycles++;
					  System.out.println("A Hamiltonian Cycle is Found");
					  for(List<Integer> cycle : cycles){
							for(int node : cycle){
								 System.out.print(node+" ");
							}
					  }
					  System.out.println("\nc Iteration : " + countOfIterations);
					  System.out.println("c Cycles : "+ countOfCycles);
					  stop = System.nanoTime();
					  solvetime = (stop-start) / 1000000000.0;
					  System.out.printf("c Solving Time : %f sec\n", solvetime);
					  System.out.printf("c Average solving Time : %f sec\n", solvetime/(countOfIterations+1));
					  System.exit(0);
				 }
				 else{
					  System.out.println("c "+cycles.size()+" cycles found");
					  countOfCycles += cycles.size();
					  for(List<Integer> cycle : cycles)
							addBlockingClauses(cycle, satFileName);
				 }
			} else break;
		}
		System.out.println("c This Graph has no Hamiltonian Cycle");
		System.out.println("c Iteration : " + countOfIterations);
		System.out.println("c Cycles : "+ countOfCycles);
		stop = System.nanoTime();
		solvetime = (stop-start) / 1000000000.0;
		System.out.printf("c Solving Time : %f sec\n", solvetime);
		System.out.printf("c Average solving Time : %f sec\n", solvetime/(countOfIterations+1));
	 }
	
}
