package iSugar;

import iSATLibrary.ISatSolver;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

import jp.kobe_u.sugar.SugarException;
import jp.kobe_u.sugar.converter.Converter;
import jp.kobe_u.sugar.converter.Simplifier;
import jp.kobe_u.sugar.csp.BooleanVariable;
import jp.kobe_u.sugar.csp.CSP;
import jp.kobe_u.sugar.csp.IntegerDomain;
import jp.kobe_u.sugar.csp.IntegerVariable;
import jp.kobe_u.sugar.encoder.Encoder;
import jp.kobe_u.sugar.encoder.Problem;
import jp.kobe_u.sugar.expression.Expression;
import jp.kobe_u.sugar.expression.Parser;
import static iSugar.Solver.Status.*;

public class Solver extends Problem {

	protected class Status{
		private Status(){}
		public static final int SATISFIABLE = 10;
		public static final int UNSATISFIABLE = 20;
		public static final int INDETERMINATE = 0;
	}

	protected static String version = "iSugar v-1.0.0";	
	protected static CSP csp;
	protected static boolean incremental = true;
	protected static boolean enumeration = false;
	protected static boolean hcp = false;
	protected static boolean isMix = false;
	protected static String problemFile;
	
	protected ISatSolver satSolver;
	protected int verbosity = 1;
	protected String solverName;
	
	public Solver(String solverName){
		this.solverName = solverName;
		this.satSolver = new ISatSolver(solverName);
	}
	
	public void setBufferSize(int limit){
		satSolver.setBufferSize(limit);
	}
	
	public int[] makeArray(){
		return satSolver.makeBuffer();
	}
	
	public void _addClause(int[] clause){
		satSolver.add_a_clause(clause);
	}
	
	public void addNormalizedClause(int[] lits) {
		satSolver.addNormalizedClause(lits);
	}
	
	public void done(){
		satSolver.done();
	}
	
	 public void add_clause(int[] clause){
		 satSolver.add_clause(clause);
	 }
	
	 public int solve(){
		 return satSolver.solve();
	 }	

	 public int solve(int[] assumption){
		 return satSolver.solve(assumption);
	 }	
	 
	 public void setTerminate(int state){
		 satSolver.ipasir_set_terminate(state);
	 }

	 public void addBlockingClause(int[] literals){
		 satSolver.add_blocking_clause(literals);
	 }
	
	public void addBlockingClause(){
		satSolver.add_blocking_clause();
	}

	public void setFrozen(int literal){
		satSolver.set_frozen(literal);
	}
	
	public boolean isEliminated(int literal){
		return satSolver.is_eliminated(literal);
	}
	
	public BitSet getModel() {
		return satSolver.get_model();
	}
	
	public int[] getIntModel(){
		return satSolver.get_int_model();
	}
	
	public void outputModel(){
		satSolver.output_model();
	}

	public double getTime() {
		return satSolver.get_time();
	}

	public String getSolverName() {
		 return satSolver.get_solver_name();
	}

	public String getSolverSignature(){
		return satSolver.ipasir_signature();
	}
	
	public void finalize() throws Throwable{
	    try {
	        super.finalize();
	      } finally {
	        satSolver.ipasir_release();
	      }
	}
	
	public void printSignature(){
		System.out.println(incremental ? version + " + " + getSolverName() : version);
	}

	public static List<Expression> parse(String cspFileName) throws SugarException, IOException {
		InputStream in;
		if (cspFileName.endsWith(".gz")) {
			in = new GZIPInputStream(new FileInputStream(cspFileName));
		} else {
			in = new FileInputStream(cspFileName);
		}
		BufferedReader reader = new BufferedReader(new InputStreamReader(in, "UTF-8"));
		Parser parser = new Parser(reader);
		List<Expression> expressions = parser.parse();
		reader.close();
		reader = null;
		in.close();
		in = null;
		parser = null;
		Runtime.getRuntime().gc();
		return expressions;
	}

	public static void loadCSP(String cspFileName) throws SugarException, IOException {
		List<Expression> expressions = parse(cspFileName);
		csp = new CSP();
		Converter converter = new Converter(csp);
		converter.convert(expressions);
		converter = null;
		expressions = null;
		Expression.clear();
		Runtime.getRuntime().gc();
		csp.propagate();
		if (csp.isUnsatisfiable()) {
			return;
		}
		Simplifier simplifier = new Simplifier(csp);
		simplifier.simplify();
		Runtime.getRuntime().gc();
	}

	 /*
	  * encoding CSP to SAT
	  */
	public void encode() throws SugarException {
		long start = 0, stop = 0;
		if (verbosity > 0){
			System.out.println("c Encoding...");
			start = System.nanoTime();
		}
		Encoder encoder = new Encoder(csp);
		encoder.encode(this);
		//satSolver.done();
		if (verbosity > 0) {
			stop = System.nanoTime();
			System.out.println("c Finished Encoding.");
			System.out.printf("c Encoding Time : %f sec\n", (stop-start) / 1000000000.0);
			System.out.printf("c variables : %d\nc clauses : %d\n", variablesCount, clausesCount);
		}
	 }

	public int solveCSP() throws SugarException{
		return solveCSP(null);
	}
	
	public int solveCSP(int assumption[]) throws SugarException {
		if (csp.isUnsatisfiable()) {
			return UNSATISFIABLE;
		}
		int solved = solve(assumption);
		//boolean solved = assumption == null ? solve()==SATISFIABLE : solve(assumption)==SATISFIABLE;
		if (solved==SATISFIABLE) {
			BitSet model = getModel();
			for (IntegerVariable v : csp.getIntegerVariables()) {
				v.decode(model);
			}
			for (BooleanVariable v : csp.getBooleanVariables()) {
				v.decode(model);
			}
		}
		return solved;
	}
	
	/*
	 * output solution
	 */
	public void outputSolution(){
		for (IntegerVariable vv : csp.getIntegerVariables())
			 if(!vv.isAux()) { System.out.println("a " + vv.getName() + "\t" + vv.getValue()); }
		System.out.println("a ");
	}

	
	/**
	 * Solves an optimization problem.
	 * @throws SugarException
	 */
	public void solveOpt() throws SugarException {
		
		boolean isMin = csp.getObjective().equals(CSP.Objective.MINIMIZE) ? true : false;

		IntegerVariable v = csp.getObjectiveVariables().get(0); // objective variable
		IntegerDomain d = v.getDomain();
		int assumption[] = new int[1];
		ArrayList<Integer> domain = new ArrayList<Integer>();
		
		Iterator<Integer> iter = d.values();
		while (iter.hasNext()) { domain.add(iter.next()); }

		int mid; // the value of assumption (p(x<=mid) or -p(x<=mid))
		int numberOfSolve = 0; 
		double solveTime, optSolveTime;
		int offset = 0; 
		int tmpopt = d.getUpperBound(); // temporal optimum
		int solved;
		boolean isSAT = false; // if one solution is found, turn to be true
		
		// searching routine
		optSolveTime = getTime();
		while (true) {

			numberOfSolve++;
			System.out.printf("c %dth Solving.\n", numberOfSolve);

			// making assumption clause
			if(isMix && domain.size() < 100){
				mid = isMin ? domain.size()-1 : 0;
			} else {
				mid = (isMin || domain.size()==1) ? domain.size() / 2 : domain.size() / 2 - 1;
			}
			assumption[0] = (isMin || domain.size()==1) ? v.getCode() + mid + offset : -(v.getCode() + mid + offset);
			
			System.out.println("c OBJECTIVE BOUND : " + domain.get(0) + " <= " + v.getName()
					+ " <= " + domain.get(domain.size() - 1));
			System.out.println("c SOLVING : " + v.getName()
					+ ((isMin || domain.size()==1) ? " <= " : " > ") + domain.get(mid) + " ...");

			// solving
			solveTime = getTime();
			solved = solveCSP(assumption);
			if (solved==SATISFIABLE) { // SAT
				isSAT = true;
				tmpopt = v.getValue();
				solveTime = getTime() - solveTime;
				System.out.println("c FOUND : "	+ v.getName()
						+ ((isMin || domain.size()==1) ? " <= " : " > ") + domain.get(mid));
				System.out.printf("c Solving Time : %f sec\n",solveTime);
				// output solution
				outputSolution();
				/*
				for (IntegerVariable vv : csp.getIntegerVariables()) {
					 if(!vv.isAux()) System.out.println("a " + vv.getName() + "\t" + vv.getValue());
				}
				System.out.println("a ");
					*/				
				// update domain
				int val = mid;
				if(isMin){
					for (int i = 0; i < mid; i++) {
						if (domain.get(i) == v.getValue()) {
							val = i;
							break;
						}
					}
					while (domain.size() > val) { domain.remove(val); }
				} else {
					if (domain.size()==1) break;
					for (int i = mid+1; i < domain.size(); i++) {
						if (domain.get(i) == v.getValue()) {
							val = i;
							break;
						}
					}
					for (int i = 0; i < val; i++){
						domain.remove(0);
						offset++;
					}
				}
				// end of search
				if(isMin ? domain.size()==0 : domain.size()==1){
					optSolveTime = getTime()-optSolveTime;
					System.out.println("c OPTIMUM : " + v.getName()	+ " = " + tmpopt);
					System.out.printf("c Number of Solves : %d\n", numberOfSolve);
					System.out.printf("c Total Solving Time : %f sec\n", optSolveTime);
					return;		
				}
				
				// add clause to SAT solver as fact
					assumption[0] = isMin ? v.getCode() + val - 1 + offset : -(v.getCode() + offset - 1);
					add_clause(assumption);					
			} else { // UNSAT
				solveTime = getTime() - solveTime;
				System.out.println("c NOT FOUND : "	+ v.getName()
						+ ((isMin || domain.size()==1) ? " <= " : " > ") + domain.get(mid));
				System.out.printf("c Solving Time : %f sec\n",solveTime);

				// update domain
				if(isMin){
					for (int i = 0; i <= mid; i++) {
						domain.remove(0);
						offset++;
					}
				} else {
					while (domain.size() > mid+1) { domain.remove(mid+1); }
				}
				// end of search
				if (isMin ? domain.size() == 0 : domain.size() == 1) {
					optSolveTime = getTime()-optSolveTime;
					if (isSAT)
						System.out.println("c OPTIMUM : " + v.getName() + " = " + tmpopt);
					else
						System.out.println("s UNSATISFIABLE");
					System.out.printf("c Number of Solves : %d\n", numberOfSolve);
					System.out.printf("c Total Solving Time : %f sec\n", optSolveTime);
					return;
				}
				
				// add clause to SAT solver as fact
				assumption[0] *= -1;
				add_clause(assumption);
			}
		}
	}

	
	/**
	 * Enumerate solutions.
	 * @throws SugarException
	 * @throws IOException
	 */
	public void solveEnum() throws SugarException, IOException{

		// temporal files(for non-incremental search)
		String satFileName = "tmp.cnf";
		String mapFileName = "tmp.map";
		String outFileName = "tmp.out";
		ProcessBuilder pb = new ProcessBuilder(solverName, satFileName, outFileName);
		Encoder encoder = new Encoder(csp);
		
		int numberOfSolve = 0;
		double solveTime = 0.0; // time of each search
		double solveTotalTime = 0.0; // total time of searches
		double totalTime = 0.0; // time of total (include encoding/decoding, adding blocking clause etc)
		ArrayList<Integer> tmpArray = new ArrayList<Integer>(); // for blocking clause
		boolean isSAT = false; // if one solution is found, turn to be true
	
		if(incremental){
			encode();
			totalTime = getTime();
		} else {
			totalTime = System.nanoTime();
			encoder.encode(satFileName);
			encoder.outputMap(mapFileName);
		}
		
		// searching routine
		while(true){
  			System.out.printf("c %dth Solving...\n", numberOfSolve+1);
  			solveTime = incremental ? getTime() : System.nanoTime();
 
  			// invoke SAT solver (non-incremental search)
  			if(!incremental){
  				pb.redirectErrorStream(true);
  				try {
  					Process p = pb.start();
  					p.waitFor();
  					System.out.println(pb.redirectInput());
  					try (BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()))) {
  						for(String line = br.readLine(); line != null; line = br.readLine()) {
  							System.out.println(line);
  						}
  					}
  				} catch (IOException | InterruptedException e) {
  					e.printStackTrace();
  				}
  			}
  			
  			// solve
  			if(incremental ? solveCSP()==SATISFIABLE : encoder.decode(outFileName)){
  				solveTime = incremental ? getTime()-solveTime : System.nanoTime()-solveTime;
  				solveTotalTime += solveTime;
  				numberOfSolve++;
  				isSAT = true;
  				if(incremental) System.out.println("s SATISFIABLE");
  				
  				// decode(incremental search)
  				if(incremental){
  					BitSet model = getModel();
  					for (IntegerVariable v : csp.getIntegerVariables()) {
  						v.decode(model);
  					}
  					for (BooleanVariable v : csp.getBooleanVariables()) {
  						v.decode(model);
  					}
  				}		
  				
  				// make blocking clause
  				for(IntegerVariable v : csp.getIntegerVariables()){
  					System.out.println("a " + v.getName() + "\t" + v.getValue());
  					if(!v.isAux()){
  						int lit;
  						if(v.getValue()!=v.getDomain().getLowerBound()){
  							lit = v.getCode() + v.getDomain().sizeLE(v.getValue())-2;
  							tmpArray.add(incremental ? -lit : lit);
  						}
  						if(v.getValue()!=v.getDomain().getUpperBound()){
  							lit = v.getCode() + v.getDomain().sizeLE(v.getValue())-1;
  							tmpArray.add(incremental ? lit : -lit);
  						}
  					}
  				}
  				System.out.println();
  				
  				int size = tmpArray.size();						
  				int[] blockingClause = new int[size];					
  				for(int i=0; i<size; i++){
  					blockingClause[i] = tmpArray.get(i);
  				}

  				// add blocking clause
  				if(incremental){
  					addBlockingClause(blockingClause);
  				} else {
  					StringBuilder sb = new StringBuilder("");
  					File file = new File(satFileName);
  					BufferedReader br = new BufferedReader(new FileReader(file));
  					Pattern p = Pattern.compile("(p cnf \\d+ )(\\d+)");
  					Matcher m;
  					String line = br.readLine();
  					m = p.matcher(line);
  					if(m.find()){
  						int clauses = Integer.parseInt(m.group(2)) + 1;
  						sb.append(m.group(1)+clauses+"\n");
  					}
  					while((line=br.readLine()) != null){
  						sb.append(line+"\n");
  					}
  					br.close();
  					FileWriter filewriter = new FileWriter(file);
  					filewriter.write(sb.toString());
  					for(int i: blockingClause){
  						filewriter.write(i+" ");
  					}
  					filewriter.write("0\n");
  					filewriter.close();
  				}
  				tmpArray.clear();
  			} else {
  				// end of search
  				solveTime = incremental ? getTime()-solveTime : System.nanoTime()-solveTime;
  				solveTotalTime += solveTime;
  				if(incremental) System.out.println("s UNSATISFIABLE");
  				System.out.println("c \nc FINISHED");
  				System.out.println(isSAT ? "c FOUND SOLUTION" : "c NOT FOUND SOLUTION");
  				System.out.printf("c Number of Solutions : %d\n", numberOfSolve);
  				System.out.printf("c Solving Time : %f sec\n", incremental ? solveTotalTime : solveTotalTime/1000000000.0);				
  				break;
  			}
		}
		totalTime = incremental ? getTime()-totalTime : System.nanoTime()-totalTime;
		System.out.printf("c Total Time : %f sec\n", incremental ? totalTime : totalTime/1000000000.0);
		new File(satFileName).delete();
		new File(mapFileName).delete();
		new File(outFileName).delete();
	}
	
	public void solveHCP() throws SugarException, IOException {
		Graph graph = new Graph(this);
		System.out.println("c Parsing Problem...");
		graph.parse(problemFile);
		graph.define();
		System.out.println("c Finished.");
		if (incremental) {
			 graph.solveHCP();
		} else {
			 graph.solveHCP(solverName);
		}
	}
	
	public void solveProblem(){
		try {
			printSignature();
			setTerminate(0);
			if(hcp) {
				solveHCP();
			} else {
				loadCSP(problemFile);
				if (enumeration){
					solveEnum();
				} else if (csp.isMaximize() || csp.isMinimize()){
					encode();
					solveOpt();
				} else {
					encode();
					int solved = solveCSP();
					switch(solved){
					case SATISFIABLE:
						System.out.println("s SATISFIABLE");
						outputSolution();
						break;
					case UNSATISFIABLE:
						System.out.println("s UNSATISFIABLE");
						break;
					case INDETERMINATE:
						System.out.println("s INDETERMINATE");
						break;
					}
				}
			}			
		} catch(Exception e) {
			e.printStackTrace();
		}

	}
		
	//debug
	/*
	public void test(){
		double time;
		int lits = 10000000;
		setBufferSize(10);
		int[][] clauses = new int[100000000/lits][lits+1];
		for(int i=0; i<100000000/lits; i++){
			for(int j=0; j<lits; j++)
				clauses[i][j] = 0;
			clauses[i][lits] = 0;
		}
		
		System.out.println("Adding...");
		time = System.nanoTime();
		for(int[] clause : clauses){
			for(int lit : clause){
				satSolver.ipasir_add(lit);
			}
		}
		time = System.nanoTime() - time;
		System.out.printf("Time : %f sec\n", time/1000000000);
		
		
		System.out.println("Adding...");
		time = System.nanoTime();
		for(int[] clause : clauses){
			satSolver.add_clause(clause);
		}
		time = System.nanoTime() - time;
		System.out.printf("Time : %f sec\n", time/1000000000);
		
		System.exit(0);
	}
*/
}
